
import datetime

try:
    from OpenSSL import SSL
except ImportError:
    SSL = None

from zope.interface import implements

from twisted.application import service
from twisted.internet import defer, reactor
from twisted.protocols import policies
from twisted.python import log, failure
from twisted.cred import portal, checkers
from twisted.mail import smtp

from epsilon import extime

from vertex import sslverify

from axiom import item, attributes, userbase, scheduler, batch

from xquotient import iquotient, exmess, mimestorage

MessageSource = batch.processor(exmess.Message)

class MailConfigurationError(RuntimeError):
    """You specified some invalid configuration.
    """

class MessageDelivery(object):
    implements(smtp.IMessageDelivery)

    def __init__(self, portal):
        self.portal = portal

    def receivedHeader(self, helo, origin, recipients):
        return "" # Maybe put something here?

    def validateFrom(self, helo, origin):
        return origin

    def validateTo(self, user):
        def cbLogin((iface, avatar, logout)):
            logout() # XXX???
            def createMIMEReceiver():
                return avatar.createMIMEReceiver(
                    u"smtp://%s@%s" % (user.dest.local, user.dest.domain))
            return createMIMEReceiver
        def ebLogin(err):
            log.msg("Failed validateTo")
            log.err(err)
            return defer.fail(smtp.SMTPBadRcpt("Denied!"))
        creds = userbase.Preauthenticated('@'.join((user.dest.local, user.dest.domain)))
        d = self.portal.login(creds, None, iquotient.IMIMEDelivery)
        d.addCallbacks(cbLogin, ebLogin)
        return d


class SafeMIMEParserWrapper(object):
    """
    Simpler wrapper around a real MIME parser which captures errors
    from lineReceived and saves them until messageDone().
    """

    failure = None

    def __init__(self, receiver):
        self.receiver = receiver


    def lineReceived(self, line):
        if self.failure is None:
            try:
                return self.receiver.lineReceived(line)
            except:
                self.failure = failure.Failure()


    def messageDone(self):
        if self.failure is not None:
            self.failure.raiseException()
        else:
            return self.receiver.messageDone()


    def __getattr__(self, name):
        # Let people get at feedStringNow and that junk, if they want.
        # We won't help them out if they try.
        return getattr(self.receiver, name)



class DeliveryAgent(item.Item, item.InstallableMixin):
    implements(iquotient.IMIMEDelivery)

    installedOn = attributes.reference()
    messageCount = attributes.integer(default=0)

    def installOn(self, other):
        super(DeliveryAgent, self).installOn(other)
        other.powerUp(self, iquotient.IMIMEDelivery)


    def createMIMEReceiver(self, source):
        today = datetime.date.today()
        fObj = self.installedOn.newFile(
            'messages',
            str(today.year),
            str(today.month),
            str(today.day),
            str(self.messageCount % 100),
            str(self.messageCount))
        self.messageCount += 1
        return SafeMIMEParserWrapper(mimestorage.MIMEMessageStorer(
            self.installedOn, fObj, source))



class DeliveryFactoryMixin(object):
    implements(smtp.IMessageDeliveryFactory)

    def getMessageDelivery(self):
        return MessageDelivery(self.portal)



class MailTransferAgent(item.Item, item.InstallableMixin, service.Service, DeliveryFactoryMixin):
    typeName = "mantissa_mta"
    schemaVersion = 1

    messageCount = attributes.integer(
        "The number of messages which have been delivered through this agent.",
        default=0)
    installedOn = attributes.reference(
        "A reference to the store or avatar which we have powered up.")

    portNumber = attributes.integer(
        "The TCP port to bind to serve SMTP.",
        default=6025)
    securePortNumber = attributes.integer(
        "The TCP port to bind to serve SMTP/SSL.",
        default=0)
    certificateFile = attributes.bytes(
        "The name of a file on disk containing a private "
        "key and certificate for use by the SMTP/SSL server.",
        default=None)

    domain = attributes.bytes(
        "The canonical name of this host.  Used when greeting SMTP clients.",
        default=None)

    # These are for the Service stuff
    parent = attributes.inmemory()
    running = attributes.inmemory()

    # A cred portal, a Twisted TCP factory and as many as two
    # IListeningPorts
    portal = attributes.inmemory()
    factory = attributes.inmemory()
    port = attributes.inmemory()
    securePort = attributes.inmemory()

    # When enabled, toss all traffic into logfiles.
    debug = False

    def activate(self):
        self.portal = None
        self.factory = None
        self.port = None
        self.securePort = None

    def installOn(self, other):
        super(MailTransferAgent, self).installOn(other)
        other.powerUp(self, service.IService)
        other.powerUp(self, smtp.IMessageDeliveryFactory)
        scheduler.IScheduler(self.store).schedule(self.store.findOrCreate(MessageSource), extime.Time())
        self.setServiceParent(other)

    def privilegedStartService(self):
        if SSL is None and self.securePortNumber is not None:
            raise MailConfigurationError(
                "No SSL support: you need to install OpenSSL to server SMTP/SSL")

        realm = portal.IRealm(self.installedOn, None)
        if realm is None:
            raise MailConfigurationError(
                "No realm: "
                "you need to install a userbase before using this service.")

        chk = checkers.ICredentialsChecker(self.installedOn, None)
        if chk is None:
            raise MailConfigurationError(
                "No checkers: "
                "you need to install a userbase before using this service.")

        self.portal = portal.Portal(realm, [chk, checkers.AllowAnonymousAccess()])
        self.factory = smtp.SMTPFactory(self.portal)
        if self.domain is not None:
            self.factory.domain = self.domain

        if self.debug:
            self.factory = policies.TrafficLoggingFactory(self.factory, 'smtp')

        if self.portNumber is not None:
            self.port = reactor.listenTCP(self.portNumber, self.factory)

        if self.securePortNumber is not None and self.certificateFile is not None:
            cert = sslverify.PrivateCertificate.loadPEM(file(self.certificateFile).read())
            certOpts = sslverify.OpenSSLCertificateOptions(
                cert.privateKey.original,
                cert.original,
                requireCertificate=False,
                method=SSL.SSLv23_METHOD)
            self.securePort = reactor.listenSSL(self.securePortNumber, self.factory, certOpts)

    def stopService(self):
        L = []
        if self.port is not None:
            L.append(defer.maybeDeferred(self.port.stopListening))
            self.port = None
        if self.securePort is not None:
            L.append(defer.maybeDeferred(self.securePort.stopListening))
            self.securePort = None
        return defer.DeferredList(L)

